package dareka;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.channels.ServerSocketChannel;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class Server {
    private static final int MAX_WAITING_TIME = 60;
    private Config config;
    private ServerSocket serverSocket;
    private ExecutorService executor = Executors.newCachedThreadPool();
    private boolean stopped = false;

    public Server(Config config) throws IOException {
        if (config == null) {
            throw new IllegalArgumentException("config must not be null");
        }

        this.config = config;

        // use channel to make it available Socket#getChannel() for non blocking
        // I/O.
        ServerSocketChannel serverCh = ServerSocketChannel.open();
        serverSocket = serverCh.socket();
    }

    /**
     * Start the server. The thread which call this method is blocked until
     * stop() is called or some errors occurred.
     */
    public void start() {
        if (stopped) {
            return;
        }

        try { // ensure cleanup
            bindServerSocket();
            acceptServerSocket();
        } finally {
            Logger.info("finalizing");
            cleanupServerSocket();
            cleanupExecutor();
        }
    }

    /**
     * Stop the server. Please call this method from another thread which called
     * start().
     */
    public synchronized void stop() {
        stopped = true;

        try {
            if (!serverSocket.isClosed()) {
                serverSocket.close();
            }
            executor.shutdown();
        } catch (IOException e) {
            Logger.error(e);
        }
    }

    private void bindServerSocket() {
        try {
            serverSocket.bind(new InetSocketAddress(
                    InetAddress.getByName(null), config.getListenPort()));
        } catch (IOException e) {
            Logger.error(e);
            stop(); 
        }
    }

    private void acceptServerSocket() {
        try {
            while (!stopped) {
                Socket client = serverSocket.accept();

                synchronized (this) { // avoid to conflict with stop()
                    if (stopped) {
                        break;
                    }

                    ConnectionProcessor worker = new ConnectionProcessor(
                            config, client);
                    executor.execute(worker);
                }
            }
        } catch (IOException e) {
            // stop() is called.
            // including AsynchronousCloseException (in NIO)
            Logger.debug(e.toString());
        }
    }

    private void cleanupServerSocket() {
        if (!serverSocket.isClosed()) {
            try {
                serverSocket.close();
            } catch (IOException e) {
                Logger.warning(e.toString());
            }
        }
    }

    private void cleanupExecutor() {
        if (!executor.isTerminated()) {
            try {
                Logger.debug("waiting for terminating threads...");
                if (executor.awaitTermination(MAX_WAITING_TIME,
                        TimeUnit.SECONDS)) {
                    Logger.debug("done");
                } else {
                    Logger.debug("timed out");
                    executor.shutdownNow();
                }
            } catch (InterruptedException e) {
                Logger.warning(e.toString());
            }
        }
    }
}
